package Core;
# Author: Alex Brasetvik < alex at brasetvik dot com >
# Author: Idar Lund < effbiai at users dot sf dot net >
# Reloadable module doing most of the work of Erwin.
# Erwin.pl acts as a proxy between the POE-interface and Core.pm
# $Id$
use strict;
use 5.14.3;
no strict 'refs';
use vars qw(%plugins %config);

use Common;
use Digest::MD5 qw(md5_base64);
use User;

use Data::Dumper;
use Storable;

use DBI;
use My::DBI qw(:all);
use DBD::SQLite;

use Log::LogLite;

use Locale::gettext;
use POSIX;
setlocale(LC_MESSAGES, "");
bindtextdomain('erwin', '/usr/share/locale/');
textdomain('erwin');

#used in var().
use constant VALUE => 0;
use constant EXPIRES => 1;
#used in hostmask functions
use constant HOSTMASK => 0;
use constant CHANNELS => 1;

=head1 NAME

Core

=head1 SYNOPSIS

use Core;

my $core = new Core;

=head1 DESCRIPTION

This module takes care of some of the core functionalities of the Erwin-bot.
Erwin.pl serves as a proxy between POE and Core.pm

Core loads and reloads configuration files, modules, plugins and regexes and
checks if they work.

It also does all the logic of doing regex matching on every incoming message
and running the mapped subroutines, as well as relaying events to the various
plugins. Additionally, it maintains multiple database connections.

=head1 FUNCTIONS

=cut


sub new
{
    my $core = shift;
    my $config = shift;

    my $log =  new Log::LogLite($config->{LogFile}, $config->{LogLevel})
        or die replace(gettext('Can\'t open logfile "$FILE": '), { FILE => $config->{LogFile} }) . "$!";

    bless {}, $core;
    $core->{_log} = \$log;
    $core->{_config} = $config; 
    $core->{_nick} = $config->{Nick};
    $core->{_plugin} = {};
    $core->{_channels} = [];
    $core->{_DBI} = {};
    $core->{_SERVER} = {};

    $core->Rehash;

    $core->dbi_connect({ driver=>'SQLite', database => 'tmp', name => 'tmp' });
    my $dbh_tmp = $core->dbh('tmp');
    if(dbi_exists($dbh_tmp, 'SELECT * FROM sqlite_master WHERE name=?', 'hostmasks')) {
        prepare_and_exec($core->dbh('tmp'), 'DELETE FROM hostmasks');
        prepare_and_exec($core->dbh('tmp'), 'DELETE FROM channels');
        prepare_and_exec($core->dbh('tmp'), 'DELETE FROM cron');
    } else {
        $core->dbh('tmp')->do('CREATE TABLE hostmasks (nick PRIMARY KEY, hostmask)');
        $core->dbh('tmp')->do('CREATE TABLE channels (nick, channel, PRIMARY KEY(nick, channel))');
        $core->dbh('tmp')->do('CREATE TABLE cron (time, module, PRIMARY KEY(time, module))');
    } 

    $core->dbi_connect({ driver=>'SQLite', database => 'erwin', name => 'erwin' });
    my $dbh = $core->dbh('erwin');
    $core->{_DBI}->{_dbh}->{'users'} = $dbh;

    unless(dbi_exists($dbh, 'SELECT * FROM sqlite_master WHERE name=?', 'variables')) {
        $dbh->do('CREATE TABLE variables (namespace, keyword, value blob, expires, PRIMARY KEY(namespace, keyword))');
        $dbh->do('CREATE TABLE users (username PRIMARY KEY, password, hostmask, flags, active)');
        $dbh->do('CREATE INDEX idx_hostmasks ON users(hostmask)');
    }

    $core->flush;

    $core->log(gettext('Core object created'), DEBUG);

    return $core;
}


=head2 modules

Returns an arrayref of defined, but not necessarily loaded, modules.

=cut

sub modules
{
    return shift->{_config}->{Modules};
}

=head2 serverinfo

Returns a hashref of variables retrieved from the IRC-server.

Currently, this is MODES, MAXCHANNELS, MAXBANS, MAXTARGETS, NICKLEN, TOPICLEN, and KICKLEN.

=cut

sub serverinfo { return shift->{_SERVER}; }

=head2 log

Returns a Log::LogLite-object.

=cut

sub log
{
    my $core = shift;
    my $string = shift;
    my $level = shift;
    chomp($string); $string =~ s/\n//g;

    if($core->config->{LogStdout}) {
        print scalar(localtime(time())) . " $string\n";
    } else {
        ${$core->{_log}}->write($string, $level);
    }

}

=head2 regexes

Returns an arrayref of regexes loaded, tested and ready to be used.
Regexes in the configuration failing sanity tests are not included.

=cut

sub regexes
{
    return shift->{_regexes};
}

=head2 config

Returns a hashref of the configuration directives.

=cut


sub config
{
    return shift->{_config};
}

=head2 addhost($who, $channel)

Adds a hostmask record for the user.

=cut

sub addhost
{
    my ($core, $hostmask, $channel) = @_;
    my $nick;

    $hostmask =~ /(.*)!(.*)/;
    ($nick, $hostmask) = ($1, $2);

    dbi_save_from_hash($core->dbh('tmp'), 'hostmasks', 'nick', { nick => $nick, hostmask => $hostmask });
    #dbi_insert_from_hash($core->dbh('tmp'), 'channels', { nick => $nick, channel => $channel });
    dbi_save_from_hash($core->dbh('tmp'), 'channels', 'nick', { nick => $nick, channel => $channel });
}

=head2 $core->getHost($nick)

Returns the complete hostmask of nick.

=cut

sub getHost
{
    my $core = shift;
    my $nick = shift;

    return dbi_exec_for_scalar($core->dbh('tmp'), "SELECT nick || '!' || hostmask FROM hostmasks WHERE lower(nick)=?", lc($nick));
}

sub getUsersInChannel
{
    my $core = shift;
    my $channel = shift;

    return dbi_exec_for_row_list($core->dbh('tmp'), 'SELECT nick FROM channels WHERE channel=?', $channel);
}

sub getWhoInChannel
{
    my $core = shift;
    my $channel = shift;

    return dbi_exec_for_lol($core->dbh('tmp'), 'SELECT channels.nick, hostmask FROm channels, hostmasks WHERE channels.nick=hostmasks.nick AND channel=?', $channel);
}

sub getUserChannels
{
    my $core = shift;
    my $nick = shift;

    return dbi_exec_for_row_list($core->dbh('tmp', 'SELECT channel FROM channels WHERE nick=?', $nick));
}

=head2 removehost($who, $channel)

    Removes a hostmask record for the user.

    $channel is optional.

=cut

sub removehost
{
    my ($core, $hostmask, $channel) = @_;
    my $nick;

    $hostmask =~ /(.*)!(.*)/;
    ($nick, $hostmask) = ($1, $2);

    my $dbh = $core->dbh('tmp');

    if($channel) {
        if(dbi_exists($dbh, 'SELECT channel FROM channels WHERE nick=? AND channel!=?', $nick, $channel)) {
            prepare_and_exec($dbh, 'DELETE FROM channels WHERE nick=? AND channel=?', $nick, $channel);
        } else {
            prepare_and_exec($dbh, 'DELETE FROM channels WHERE nick=?', $nick);
            prepare_and_exec($dbh, 'DELETE FROM hostmasks WHERE nick=?', $nick);
        }
    } else {
        prepare_and_exec($dbh, 'DELETE FROM channels WHERE nick=?', $nick);
        prepare_and_exec($dbh, 'DELETE FROM hostmasks WHERE nick=?', $nick);
    }
}

=head2 RehashModules

This function reloads all the modules.
Their corresponding records in %INC are deleted to force a read from
disk. If a module isn't successfully reloaded, the errors will be trapped, but
the module in question will no longer work, so it's always a good idea to run
Erwin with the --try-flag before provoking a reload of modules!

=cut

sub RehashModules
{
    my $core = shift;

    $core->log(gettext('Rehashing modules.'), INFO);

    foreach my $module (keys %{$core->modules}) {
        #First we eval to catch any exceptions.
        eval {
            no warnings;
            #Delete from INC to force reload from disk.
            delete $INC{"$module.pm"};
            require "$module.pm";

            my $init = "${module}::init";
            eval { &{$init}($core); };

        };

        if ($@) {
            #Loading didn't succeed.
            my $error = replace(gettext('Error loading $MODULE. Module disabled.'), { MODULE => $module });
            $core->log($error, ERROR);
            print "$@\n" if($core->config->{LogStdout});
        } else {
            if($plugins{$module} && !$core->{first_run}->{$module}) {
                print "Trying ${module}::first_run\n";
                eval { &{"${module}::first_run"}($core); };
                if($@) {
                    print "*** ERROR LOADING $module: \n", $@;
                    next;
                }
                $core->{first_run}->{$module} = 1 unless $@;
            }
            $core->log(replace(gettext('Module $MODULE loaded.'), { MODULE => $module }), INFO);
        }

    }

    $core->log(gettext('Rehashed modules.'), INFO);
}

=head2 RehashRegexes

$core->RehashRegexes will reload the regexes used. Note that sanity checks are
performed on the regexes, and that failing regexes will not be used.

=cut

sub RehashRegexes
{
    my $core = shift;

    $core->log(gettext('Reloading regexes.'), INFO);
    $core->{_testing} = 1;

    #Note: $core->config->{Regexes} is the _defined_ regexes.
    #      $core->regexes is the ones in *use* - i.e. those that have been verified and tested.

    my @oldregexes = @{$core->config->{Regexes}};

    my @regexes;

    my %regexes; #For "signatures".

        foreach my $regex (@oldregexes) {
            my %regex = %$regex;

            #Check that it works at all.
            eval { /$regex{regex}/; };
            if($@) {
                $core->log(replace(gettext('The regex "$REGEX" is invalid.'), { REGEX => $regex{regex} }), ERROR);
                print $@ if($core->config->{LogStdout});
                next; #Don't use it.
            }

            #It works. If it has a sub -- test it.
            if($regex{sub}) {

                #Hand off an empty arrayref as backrefs and an empty hashref as arguments.
                eval { &{$regex{sub}}($core, [], {}); };

                if($@) {
                    $core->log(replace(
                                gettext('The regex "$REGEX" is mapped to the sub "$SUB", but "$SUB" doesn\'t exist or work.'),
                                { REGEX => $regex{regex}, SUB => $regex{sub} }), ERROR);
                    print $@ if($core->config->{LogStdout});
                    next; #Don't use it.
                }
            }

            $regex{args} ||= {};
            my $sig = join " ", map({ $regex{$_} } qw(regex pointer content sub)), values(%{ $regex{args} });

            next if($regexes{$sig});

            $regexes{$sig} = 1;
            push(@regexes, {
                    regex => $regex{regex},
                    pointer => $regex{pointer},
                    content => $regex{content},
                    flags => $regex{flags},
                    nohalt => $regex{nohalt},
                    sub => $regex{sub},
                    args => $regex{args},
                    privmsg => $regex{privmsg},
                    public => $regex{public},
                    });

        }

    $core->{_regexes} = \@regexes;
    $core->{_testing} = 0;

    return;
}

=head2 testing

Returns a boolean.

True if Core is in the process of (re)loading modules and doing sanity checks.

All subroutines mapped by a regex should test its conditions and return immediately if true.

=cut

sub testing
{
    return shift->{_testing};
}

=head2 Rehash

$core->Rehash will reload users, modules and regexes.
If a module isn't successfully reloaded, the errors will be trapped, but
the module in question will no longer work, so it's always a good idea to run
Erwin with the --try-flag before provoking a reload of modules!

=cut

sub Rehash
{
    my $core = shift;

    no warnings;
    #Delete from INC to force reload from disk.
    $core->log(gettext('Reloading configuration files.'));

    foreach my $config (@{$core->config->{ConfigFiles}}) {

        #All configuration files must work for configuration to be reloaded.
        delete $INC{$config}; #Reload from disk.
            eval { require $config; };

        if($@) {
            $core->log(replace(gettext('Couldn\'t reload configuration from $FILE.'), { FILE => $config }), ERROR);
            $core->log(replace(gettext('Not rehashing configuration files.'), { FILE => $config }), ERROR);
            return 0;
        }
    }

    my @configs = @{$core->config->{ConfigFiles}};

    #foreach my $config (@{$core->config->{ConfigFiles}}) {   
    foreach my $config (@configs) {
        $core->log(replace(gettext('Reloading configuration file "$FILE"'), { FILE => $config }), INFO);
        $core->load_config($config);
    }

    $core->RehashModules;
    $core->RehashRegexes;

    return 1;
}

=head2 Quit

Disconnects from the server if given an argument - a quit-message.

Example:

$core->Quit($kernel, 'Go die');

=cut


sub Quit
{
    my ($core, $message) = @_;

    $core->log(gettext('Disconnecting.'), INFO);

    $core->erwin->yield(quit => ($message) ? $message : 'Leaving.');

    $core->log(gettext('Shutting down.'), INFO);
    $core->erwin->yield('shutdown');
}

=head2 cron

$core->cron is run every now and then.

=cut

sub cron
{
    my $core = shift;
    #printf "%s, %s\n", time(),  $core->{cron}->{queue}->[0][0] if($core->{cron}->{queue}->[0][1] ne 'Core');

    my $t = time();
    $core->{cron}->{core} ||= $t;

    if(time() - $core->{cron}->{core} > $core->config->{Modules}->{Core}->{cron_interval}) {
        $core->flush;
        $core->{cron}->{core} = $t;
    }

    my @queue = dbi_exec_for_lol($core->dbh('tmp'), "SELECT * FROM cron WHERE time <= ? ORDER BY time ASC", $t);

    return unless @queue;

    #while(@queue && $queue[0][0] <= $t) {
    while(@queue && $queue[0][0] <= $t) {
        my ($time, $caller) = @{ shift @queue };

        my $subref = \&{"${caller}::cron"};
        eval { &$subref($core); };
        prepare_and_exec($core->dbh('tmp'), "DELETE FROM cron WHERE time=? AND module=?", $time, $caller);
    }

=cut

        return unless @{$core->{cron}->{queue}};

    while ($core->{cron}->{queue}->[0][0] && ($core->{cron}->{queue}->[0][0] <= $t)) {
        print Dumper($core->{cron}->{queue});

        my($time, $caller) = @{ shift @{ $core->{cron}->{queue} } };

        #printf "\@ %s (%s)\n", time(), $caller;
        #Some class has ->addTick()-ed.
        my $subref = \&{"${caller}::cron"};
        eval { &$subref($core); };

    }

=cut

}

sub removeAllTicks
{
    my $core = shift;

    my $module = scalar(caller());

    prepare_and_exec($core->dbh('tmp'), 'DELETE FROM cron WHERE module = ?', $module);
}

=item addTick

Add a tick to the queue.

$core->addTick(time()+60, ['Module']);

=cut

sub addTick
{
    my ($core, $n, $module, $keep) = @_;

    return unless($n !~ /\D/ && $n > 0);
    my $t = time();
    $n += $t if($n < $t);

    $module ||= scalar(caller());

    unless(dbi_exists($core->dbh('tmp'), 'SELECT time FROM cron WHERE time=? AND module=?', $n, $module) ) {
        dbi_insert_from_hash($core->dbh('tmp'), 'cron', { time => $n, module => $module } );
    }

#   prepare_and_exec unless($keep);

=cut
    my @queue = @{ $core->{cron}->{queue} };
    push(@queue, [$n, $module]);
    @queue = sort { $a->[0] <=> $b->[0] } @queue;

    $core->{cron}->{queue} = \@queue;
=cut

}

=head2 var

$core->var is used to store (semi)persistant variables.

Example:

my $variable = $core->var({
        key => "my_key",		#Mandatory.
        value => $myref,		#Set or update the value.
        namespace => "daily",	#Default is "variables".
        expires => 600,		#Time in seconds.
        }); #Returns the correspondig value.

Default namespace is "variables", to which no magic apply.
The "expires"-namespace is reserved.

If an expire time is given, the variable will be purged afther the given number of seconds.

=cut

sub var
{
    my $core = shift;
    my $arg = shift;

    return 0 unless($arg->{key} || $arg->{delete});
    $arg->{keyword} = $arg->{key};

    my $dbh = $core->dbh('erwin');

    $arg->{namespace} ||= 'variables';

    #Expire variables after an hour by default.
    $arg->{expires} ||= 3600;
    $arg->{expires} = 3600 if($arg->{expires} !~ /^[\-\d]$/);

    if($arg->{delete}) {
        prepare_and_exec($dbh, 'DELETE FROM variables WHERE namespace=? AND keyword=?', $arg->{namespace}, $arg->{keyword});
        return 0;
    }

    my %var = dbi_exec_for_row_hash($dbh, 'SELECT * FROM variables WHERE namespace=? AND keyword=?', $arg->{namespace}, $arg->{keyword});

    if ($arg->{value}) {
    #A value is given, so set/update it.

        if($arg->{expires} > 0) {
            $arg->{expires} += time();
        }

        if($var{value}) {
            prepare_and_exec($dbh, 'UPDATE variables SET value=? WHERE namespace=? AND keyword=?', $arg->{value}, $arg->{namespace}, $arg->{keyword});
        } else {
            dbi_insert_from_hash($dbh, 'variables', { namespace => $arg->{namespace}, keyword => $arg->{keyword},
                    value => $arg->{value}, expires => $arg->{expires} });
        }
    } else {
        if($var{expires} < time()) {
            prepare_and_exec($dbh, "DELETE FROM variables WHERE namespace=? AND keyword=?", $arg->{namespace}, $arg->{keyword});
            return undef;
        }
    }
    return $var{value};
}

sub variables { return shift->{_variables}; }

=head2 upcounter($core, { namespace => $ns, key => $key, delta => -1 })

Increments the value with $delta and returns the new value.

Default delta is 1.

=cut

sub upcounter
{
    my $core = shift;
    my $arg = shift;

    return unless($arg->{key});
    $arg->{namespace} ||= 'variables';
    $arg->{delta} ||= 1;

    my $dbh = $core->dbh('erwin');

    my %var = dbi_exec_for_row_hash($dbh, 'SELECT * FROM variables WHERE namespace=? AND keyword=?', $arg->{namespace}, $arg->{key});

    if($var{expires} > 0 && $var{expires} < time()) {
        prepare_and_exec($dbh, 'DELETE FROM variables WHERE namespace=? AND keyword=?', $arg->{namespace}, $arg->{key});
    }

    $arg->{expires} ||= time() + 3600;

    if($var{value}) {
        prepare_and_exec($dbh, 'UPDATE variables SET value += ? WHERE namespace=? AND keyword=?', $arg->{delta}, $arg->{namespace}, $arg->{key});
        return $var{value} + $arg->{delta};
    } else {
        dbi_insert_from_hash($dbh, 'variables', { namespace => $arg->{namespace}, keyword => $arg->{key}, value => $arg->{delta}, expires => $arg->{expires} });
        return $arg->{delta};
    }
}

=head2 flush

Deletes expired variables.

=cut

sub flush
{
    my $core = shift;

    prepare_and_exec($core->dbh('erwin'), 'DELETE FROM variables WHERE expires > 0 AND expires < ?', time());
}

=head2 reset

Delete event specific values from heap.

=cut

sub reset
{
    my $core = shift;

#    $core->log('reset(): ' . gettext('Resetting heap.'), DEBUG);

    foreach my $var ( qw(who where msg reply nick action redirect keyword floodoverride literal privmsg public) ) {
        $core->heap->{$var} = undef;
    }
}

=head2 plugins

Returns a hash ref of available plugins and their configuration.

=cut

sub plugins { return \%plugins; }

=head2 plugin

Returns a hash ref of events with the value of the current plugin in charge.

=cut

sub plugin { return shift->{_plugin}; }

=head2 message

Run for every incoming message. 

Relays events to plugins and subroutines mapped by regexes.

Takes care of anti flood.

=cut

sub message
{
    my $core = shift;
    my $who = $core->heap->{who};
    my $msg = $core->heap->{msg};

    my ($where, %config);
    if($core->heap->{public}) {
        $where = $core->heap->{where};
        $core->module_config(\%config, $where);
    } else {
        $core->module_config(\%config, 'privmsg');
    }

    #FiSH #channel decryption
    my $fish_key = $core->config->{'Channel_settings'}->{$where}->{'fish_key'} if $core->config->{'Channel_settings'}->{$where}->{'fish_key'};
    my $fish_regex = "\\". $core->config->{'Channel_settings'}->{$where}->{'fish_prefix'} if $core->config->{'Channel_settings'}->{$where}->{'fish_prefix'};
    if ($fish_key && $msg =~ /^$fish_regex(.+)$/) {
        my $enc_msg = $1;
        $core->log("got FiSH encrypted message on $where, decrypting");
        require Crypt::ircBlowfish;
        my $blowfish = new Crypt::ircBlowfish;
        $blowfish->set_key($fish_key);
        $msg=$blowfish->decrypt($enc_msg);
    }

PLUGIN: foreach my $plugin (keys %plugins) {
            my $conf = $plugins{$plugin};

            my $type;
            #If a plugin wants charge over incoming messages, give charge to the correct plugin.
            #Skip it if it shouln't be used.
            if($core->heap->{public}) {
                next PLUGIN unless($conf->{events}->{public});
                if($core->plugin->{public}->{$where}) {
                    next PLUGIN unless($core->plugin->{public}->{$where} eq $plugin || $conf->{immutable});
                }
                $type='public';
            } else {
                next PLUGIN unless($conf->{events}->{privmsg});
                if($core->plugin->{privmsg}->{$who}) {
                    next PLUGIN unless($core->plugin->{privmsg}->{$who} eq $plugin || $conf->{immutable});
                }
                $type='privmsg';
            }

            my $sub = "${plugin}::$type";
            my $subref = \&{$sub};

            if($type eq 'public') {
                eval { &$subref($core, $who, $core->heap->{where}, $msg); };
            } else {
                eval { &$subref($core, $who, $msg); };
            }

            $core->log($@, ERROR) if $@;

        } #foreach

        return if($core->heap->{nick} eq $core->mynick); #Regexes shouldn't be run on own messages. (Although plugins should.)

            if($core->heap->{privmsg} && $core->{conversation}->{$core->heap->{nick}}) {
                #We're having a conversation.
                return if $core->conversate($msg);
            }

        #Disregard regexes if a plugin is in charge.
        return if(($core->plugin->{privmsg}->{$who} && $core->heap->{privmsg}) || ($core->plugin->{public}->{$where} && $core->heap->{public}));

        my @queue; #Queue of messages to be posted.

        #Every regex might include some keywords which should be substituted.
        my $nick = quotemeta($core->mynick);
        my $mynick = "$nick(?:\\:|\\,|\\s+)(?:\\s+)?";
        my $nicklen = $core->serverinfo->{NICKLEN}; $nicklen ||= 9;
        my $somenick = "(?:[^\s\$]{1,$nicklen}(?:\\:|\\,))(?:\\s+)?";

        # Iterate through every regex for each message recieved.
        REGEX: foreach my $regexref (@{$core->regexes}) {
            my $output;

            my %regex = %$regexref;
            my $regex = $regex{regex};

            #Don't do anything if the regex is intended for private and sent in public or vice versa.
           next REGEX if( ($core->heap->{public} && $regex{privmsg}) || ($core->heap->{privmsg} && $regex{public}) );

            $regex =~ s/\$NICK/$nick/g;
            $regex =~ s/\$MYNICK(?:,|\:|\s+)?(?:\s+)?/$mynick/g;
            $regex =~ s/\$MYNICK/$mynick/g;
            $regex =~ s/\$SOMENICK(?:,|\:|\s+)?(?:\s+)?/$somenick/g;
            $regex =~ s/\$SOMENICK/$somenick/g;

            #      $core->log('message(): ' . replace(gettext('Trying /$REGEX/'), { REGEX => $regex }), DEBUG);

            if ($msg =~ /$regex/i) {
#A hit! :-) Let's do something useful. (or at least try)
                $core->log('message(): ' . replace(gettext('$MSG matches /$REGEX/.'), { MSG => $msg, REGEX => $regex }), DEBUG);

                my $sub = $regex{sub};

                if ($sub) {
                    $core->log('message(): ' . replace(gettext('/$REGEX/ is mapped to $SUB.'), { REGEX => $regex, SUB => $sub }), DEBUG);
                    # The regex that matched has a subroutine attached to it. Execute it if the conditions are satisfactory.
                    # Check if necessary privileges are granted. If the subroutine has a flag attached to it, check if the user has it in his.

                    my $user = User->GetByHostmask($core->dbh('erwin'), $core->heap->{who});
                    my $userflags = $user->flags if($user);

#Has he/she got privileges?
                    if($regex{flags}) {
                        $core->log('message(): ' . replace(gettext('$SUB requires one of the following flags: $FLAGS.'),
                                   { SUB => $regex{sub}, FLAGS => $regex{flags} }), DEBUG);

                       next REGEX unless($userflags);

                       my $granted = 0;
                       foreach my $flag (split(//, $userflags)) {
                           $granted++ if(index($regex{flags}, $flag) > -1);
                       }

                       if ($granted) {
                           $core->log('message(): ' . replace(gettext('$USER is granted access to $SUB'), { USER => $who, SUB => $sub }), DEBUG);
                       } else {
                           $core->log('message(): ' . replace(gettext('$USER is not allowed to use $SUB'), { USER => $who, SUB => $sub }), DEBUG);
                           next REGEX;
                        }
                   }

                    #$1, $2, etc. is now the backreferences to the regex that matched $msg.
                    #Store these in @backrefs.
                    my(@backrefs);
                    push(@backrefs, ${$_}) foreach(1..$#+); # $#+ is the last backreference.

                    $core->log('message(): ' . replace(gettext('$REGEX produced the following backrefs:'), { REGEX => $regex }) . join(', ', @backrefs), DEBUG);

                    # This references an anonymous subroutine as defined by the regex,
                    # given that it's found in the list of available subroutines.
                    my $coderef = \&{$regex{sub}};

                    my $args;
                    if(ref($regex{args}) eq 'HASH') {
                        $args = $regex{args};
                    } else {
                        $args = {};
                    }

                    # Execute the sub, but trap any errors.
                    eval {
                        $output = &$coderef($core, \@backrefs, $args);
                    };

                    if($@) {
                        my $error = $@;
                        $error =~ s/\n/ /g;
                        $core->log('message(): ' . replace(gettext('$SUB returned errors: '), { SUB => $sub }) . $error, ERROR);
                    }

                } elsif ($regex{content}) {
                    # The regex has some witty answer not generated by any sub.

                    $core->log('message(): ' . replace(gettext('$REGEX is mapped to content: '), { REGEX => $regex }) . $regex{content}, DEBUG);

                    my @content = split(/\|\|\|/, $regex{content});

#Uh.. What?
                    my $keyword = $1; #$1 from the $string =~ far up there. :)
                    $output = random_string(@content);
                    $output =~ s/\$keyword/$keyword/ig;
#/Uh.

                } elsif ($regex{pointer}) {
                    if($core->modules->{Facts}) {
                        # The regex has a reference to some witty answer available in the fact-database.
                        $core->log('message(): ' . replace(gettext('$REGEX is mapped to a factoid with keyword "$KEYWORD" '),
                        { REGEX => $regex, KEYWORD => $regex{pointer} }), DEBUG);

                        eval {
                            $output = &Facts::Lookup($core, [$regex{pointer}], {});
                        }
                    } else {
                        # For some reason, it has no sub, content or pointer, so we next() without halting.
                        next;
                    }
                }
            } #if($msg =~ /$regex/)

            next REGEX unless($output =~ /\S/);
            #Yay. Let's say something :D

            chomp($output);
            push(@queue, $output);
            last REGEX unless($regex{nohalt});

        } #REGEX: 

        foreach my $msg (@queue) {
            $core->msg($core->heap->{where}, $msg);
        }
}

=head2 msg($core, $where, $text, $args)

    Send a message.

    If $args->{notice} is true, a NOTICE is sent instead of the default PRIVMSG.
    If $args->{flood} is true, antiflood mechanisms are ignored.

=cut

sub msg
{
    my ($core, $where, $text, $args) = @_;

    unless($where) {
        $core->log("No \$where in call from " . scalar(caller()), DEBUG);
        return;
    }

    my %config;
    if($where =~ /^(#|&)/) {
        $core->module_config(\%config, $where);
    } else {
        $core->module_config(\%config, 'privmsg');
    }

    #Do not say the same thing more than once every <antiflood> seconds unless forced to.
    if($config{Antiflood} > 0 && !( $core->heap->{floodoverride} || $args->{flood} ) ) {

        my $last_hit = $core->var({
                key       => md5_base64( $where . $text ),
                namespace => 'antiflood'
                });

        my $x = time() - $last_hit;

        if ( !$last_hit || ( $x > $config{Antiflood})) {
            $core->var({
                    key       => md5_base64( $where . $text ),
                    namespace => 'antiflood',
                    value     => time()
                    });
        } else {
            $core->log('message(): ' . replace(gettext('Would\'ve said "$OUTPUT" to "$WHERE", but I\'ve already done so the last $SEC seconds'),
                        { OUTPUT => $text, WHERE => $where, SEC =>$config{Antiflood} }), DEBUG);
            return;
        }
    }
    #Various substitutions, etc.:
    unless($core->heap->{literal}) {
        if($text =~ /\|\|\|/) { #Split on |||-s and return one of them.
            my @replies = split(/\|\|\|/, $text);
            $text = random_string(@replies);
        }

        my $nick = $core->heap->{nick};
        my $keyword = $core->heap->{keyword};

        if($core->heap->{redirect}) {
            $nick = $core->heap->{redirect};
            $nick =~ /(\S*)/; #Remove stupid whitespace, if any.
                $nick = $1;
            if($text =~ /\$who/i) {
                $text =~ s/\$who/$nick/gi;
            } else {
                $text = "$nick: " . $text;
            }
        } else {
            $text =~ s/\$who/$nick/gi;
        }

        $text =~ s/\$(X|KEY)/$keyword/gi;

        my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
        $mon++;

        $sec = "0$sec" if($sec < 10);
        $min = "0$min" if($min < 10);
        $hour = "0$hour" if($hour < 10);
        $mday = "0$mday" if($mday < 10);

        my @months = ('', gettext('January'), gettext('February'), gettext('March'), gettext('April'), gettext('May'), gettext('June'),
                gettext('July'), gettext('August'), gettext('September'), gettext('October'), gettext('November'), gettext('December'));

        my $monthname = $months[$mon];
        $mon = "0$mon" if($mon < 10);

        $year += 1900;

        $text =~ s/\$MDAY/$mday/g;
        $text =~ s/\$MONTH/$mon/g;
        $text =~ s/\$MNAME/$monthname/g;
        $text =~ s/\$YEAR/$year/g;
        $text =~ s/\$HOUR/$hour/g;
        $text =~ s/\$MIN/$min/g;
        $text =~ s/\$SEC/$sec/g;

        if($text =~ /RAND\((\d+?)\)/) {
            my $rand = int(rand($1));
            $text =~ s/RAND\($1\)/$rand/;
        }
    } #unless literal

    my $type = $args->{notice} ? 'notice' : 'privmsg';

    $core->log('message(): ' . replace(gettext('Posting "$MSG" to "$WHERE" ($TYPE)'), { MSG => $text, WHERE => $where, TYPE => $type }), DEBUG);

    #Check if the $text has several lines, and split them into an array
    #need to encrypt if the channels uses FiSH aswell
    my ($blowfish, $fish_key, $fish_prefix);
    if ($core->config->{'Channel_settings'}->{$where}->{'fish_key'}) {
        require Crypt::ircBlowfish;
        $fish_key = $core->config->{'Channel_settings'}->{$where}->{'fish_key'};
        $fish_prefix = $core->config->{'Channel_settings'}->{$where}->{'fish_prefix'};
        $blowfish =new Crypt::ircBlowfish;
        $blowfish->set_key($fish_key);

    }

    #  use IRC::Utils qw(GREEN NORMAL);
    #  $text = GREEN."$text".NORMAL."normal";
    #  my $hex_lol = hex (0x3);
    #  `echo text: $text > /home/tor/testfil`;
    #  `echo lol: $hex_lol >> /home/tor/testfil`;

    my @several_lines;
    if($text =~ /\n|\r/) {
        @several_lines = split(/\n|\r/, $text);
    }

    if($text =~ /<action>(.*)/ && !$core->heap->{literal}) {
        return unless($1);
        $core->erwin->yield(ctcp => $where => "ACTION $1");
    } elsif($text =~ /<reply>(.*)/ && !$core->heap->{literal}) {
        next unless($1);
        $text = $1;
        $core->erwin->yield($type => $where => $text );
    } elsif (scalar @several_lines > 1 && !$core->heap->{literal}) {
        $core->log("using several line method", DEBUG);
        foreach my $line (@several_lines) {
            $line = "$fish_prefix" . $blowfish->encrypt($line) if $blowfish;
            $core->erwin->yield($type => $where => $line );
        }
    } else {
        $core->log("using one line method", DEBUG);
        $text = "$fish_prefix" . $blowfish->encrypt($text) if $blowfish;
        $core->erwin->yield($type => $where => $text );
    }

    #We have to make a new event manually:

    $core->heap->{where} = $where;
    $core->heap->{who} = $core->getHost($core->mynick);
    $core->heap->{nick} = $core->mynick;
    $core->heap->{msg} = $text;

    if($where =~ /^#/) {
        $core->heap->{public} = 1;
        $core->heap->{privmsg} = 0;
    } else {
        $core->heap->{public} = 0;
        $core->heap->{privmsg} = 1;
    }

    $core->message;

}

=head2 mode($core,$bywhom,$where,$mode,$arg)

    Relays mode changes to various plugins.

    Note: Multiple modes such as "+mvv nick1 nick2" is not supported yet.

=cut

# XXX - Multiple modes not working!
sub mode
{
    my ($core,$bywhom,$where,$mode,@args) = @_;

    #    $core->log(replace(gettext('Modechange on $CHANNEL by $BYWHOM: $MODE $ARG'),
    #       { BYWHOM => $bywhom, MODE => $mode, ARG => $arg, CHANNEL => $where }), INFO);

    foreach my $plugin (keys %plugins) {
        my $conf = $plugins{$plugin};

    # XXX - Should plugins be able to take over modes?
    # I can't think of a practical use of it. --Alex

        next unless($conf->{events}->{mode});

        my $subref = \&{"${plugin}::mode"};
        eval { &$subref($core,$bywhom,$where,$mode,@args); };
    }

    return;
}

sub setMode
{
    my ($core, $where, $mode, $args) = @_;
    $core->erwin->yield(mode => $where => "$mode $args");
}

=head2 kick($core,$kicker,$where,$kicked)

    Relays kick events to plugins.

=cut

sub kick
{
    my ($core,$kicker,$where,$kicked, $kickmsg) = @_;
    my $hostmask = $core->heap->{users}->{$kicked};
    $core->log(replace(gettext('$KICKER kicked $KICKED from "$CHANNEL"'), { KICKER => $kicked, KICKED => $kicked, CHANNEL => $where }), INFO);

    foreach my $plugin (keys %plugins) {
        my $conf = $plugins{$plugin};

        # XXX - Should plugins be able to take over kicks?
        # I can't think of a practical use of it. --Alex

        next unless($conf->{events}->{kick});

        my $subref = \&{"${plugin}::kick"};
        eval { &$subref($core,$kicker, $where, $kicked, $kickmsg); };
    }
}

=head2 join($core, $who, $where)

    Relays join events to plugins.

=cut

sub join
{
    my ($core, $who, $where) = @_;

    my ($nick, $hostmask) = split(/!/, $who);
    $core->log(replace(gettext('$HOSTMASK joined "$CHANNEL"'), { HOSTMASK => "$nick!$hostmask", CHANNEL => $where }), INFO);

    foreach my $plugin (keys %plugins) {
        my $conf = $plugins{$plugin};

        # XXX - Should plugins be able to take over joins?
        # I can't think of a practical use of it. --Alex

        next unless($conf->{events}->{join});

        my $subref = \&{"${plugin}::join"};
        eval { &$subref($core,$who,$where); };
    }
}

=head2 part($core, $who, $where)

    Relays part events to plugins.

=cut

sub part
{
    my ($core, $who, $where) = @_;

    my ($nick, $hostmask) = split(/!/, $who);

    foreach my $plugin (keys %plugins) {
        my $conf = $plugins{$plugin};

        my $subref = \&{"${plugin}::part"};
        eval { &$subref($core,$who,$where); };
    }
}
=head2 who($core, $who, $where)

    Relays who events to plugins.

=cut

sub who
{
    my ($core, $who, $where) = @_;

    foreach my $plugin (keys %plugins) {
        my $conf = $plugins{$plugin};
        my $subref = \&{"${plugin}::who"};
        eval { &$subref($core,$who,$where); };
    }
}


=head2 quit($core, $who, $ms)

    Relays quit events to plugins.

=cut

sub quit
{
    my ($core, $who, $msg) = @_;

    my ($nick, $hostmask) = split(/!/, $who);
    $core->log(replace(gettext('$HOSTMASK quit: $MSG'), { HOSTMASK => $who, MSG => $msg }), INFO);

    foreach my $plugin (keys %plugins) {
        my $conf = $plugins{$plugin};

        next unless($conf->{events}->{quit});

        my $subref = \&{"${plugin}::quit"};
        eval { &$subref($core,$who,$msg); };
    }
}


=head2 nick($core, $who, $newnick)

Relays nick change events to plugins.

=cut

sub nick
{
    my ($core, $who, $newnick) = @_;

    my ($nick, $hostmask) = split(/!/, $who);
    $core->log(replace(gettext('$WHO changed nick to $NEWNICK'), { WHO => $who, NEWNICK => $newnick }), INFO);

    eval {
        prepare_and_exec($core->dbh('tmp'), "UPDATE hostmasks SET nick=? WHERE nick=?", $newnick, $nick);
        prepare_and_exec($core->dbh('tmp'), "UPDATE channels SET nick=? WHERE nick=?", $newnick, $nick);
    };
    print $@;

    foreach my $plugin (keys %plugins) {
        my $conf = $plugins{$plugin};

    # XXX - Should plugins be able to take over nick changes?
    # I can't think of a practical use of it. --Alex

        next unless($conf->{events}->{nick});

        my $subref = \&{"${plugin}::nick"};
        eval { &$subref($core,$who,$newnick); };
    }
}

#Obsolete, no?

sub error
{
    my $core = shift;
    $core->heap->{floodoverride} = 1; #Avoid antiflood
        return random_string(@_);
}

=head2 Join($core, $channel)

Takes a channel name as argument. TODO: Passwords

Causes the bot to join the channel.

=cut

sub Join {
    my ($core, $backrefs) = @_;
    return if($core->testing);

    my $channel = $backrefs->[0];
    $core->erwin->yield(join => $channel);
    $core->log(replace(gettext('Joined channel "$CHANNEL"'), { CHANNEL => $channel }), INFO);

    for(@{$core->channels}) {
        if($_ eq $channel) {
            return;
        }
    }

    push(@{$core->channels}, $channel);
    return;
}

sub channels { return shift->{_channels}; }

=head2 Part($core, $channel)

Takes a channel name as argument.

Causes the bot to leave the channel.

=cut

sub Part {
    my ($core, $backrefs) = @_;
    return if($core->testing);

    my $channel = $backrefs->[0];

    $core->erwin->yield(part => $channel);
    $core->log(replace(gettext('Left channel "$CHANNEL"'), { CHANNEL => $channel }), INFO);

    my @oldchannels = @{ $core->channels };
    my @channels;

    for(@oldchannels) {
        next if($_ eq $channel);
        push(@channels, $_);
    }

    $core->{_channels} = \@channels;

    return;
}

=head2 Nick($core,$newnick)

Changes the bot's nickname.

=cut

sub Nick
{
    my $core = shift;
    return if($core->testing);
    my $backrefs = shift;

    my $newnick = $backrefs->[0];

    $core->{_nick} = $newnick;
    $core->erwin->yield(nick => $newnick);
}

=head2 mynick

Returns current nick.

=cut

sub mynick { return shift->{_nick}; }


sub setKernel
{
    my ($core, $kernel) = @_;
    $core->{_kernel} = $kernel;
}

sub kernel { return shift->{_kernel} };

sub setErwin
{
    my ($core, $erwin) = @_;
    $core->{_erwin} = $erwin;
}

sub erwin { return shift->{_erwin} };

sub setHeap
{
    my ($core, $heap) = @_;
    $core->{_heap} = $heap;
}

=head2 heap

Returns a hash reference for variables set and reset for every event.

who: The hostmask to the user triggering the event.

nick: The nick part of the hostmask.

where: Where the trigger happened. This equals to the bot's nick if it's a privmsg.

msg: The unmodified string of the message.

action: Boolean. If a reply should be sent as a /me-style action.

redirect: Prefix the reply with "prefix: "

keyword: A keyword looked up by the fact module.

floodoverride: Boolean. Ignore flood protection.

literal: Boolean. Don't substitute variables.

privmsg: Boolean. True if sent as a privmsg to the bot.

public: Boolean. True if sent to a channel.

testing: Boolean. True if RehashRegexes is doing tests.

=cut


sub heap
{ 
    my $core = shift;

    if($core->{_heap}) {
        return $core->{_heap};
    } else {
        return {};
    }
}

=head2 dbi_connect

Connect to a PostgreSQL or MySQL database and store the database handler.

$core->dbi_connect({ database => ...,	#Name of database
        username => ...,	#Username
        password => ...,	#Password
        name => ...,	#Defaults to "dbh",
        driver => ...,	#Either 'Pg' or 'mysql'. Defaults to 'Pg'.
        options => { DBI-options... }
        });

Returns the database handler, which may also be accessed using B<dbh>()

    Multiple connections may be maintained. Use different names to achieve it.

    Default DBI-options are: PrintError => 0, RaiseError => 0, AutoCommit => 1

See also B<dbh>()

=cut

sub dbi_connect
{
    my $core = shift;
    my $arg = shift;

    $arg->{driver} ||= 'Pg';

    return unless($arg->{database} && $arg->{driver} =~ /^(Pg|mysql|SQLite)$/);

    unless($core->{_DBI}->{$arg->{driver}}->{DBD_LOADED}) {
        $core->log(replace(gettext('Loading DBD::$DRIVER'), { DRIVER => $arg->{driver} }), DEBUG);
        if($arg->{driver} eq 'Pg') {
            eval { require DBD::Pg; };
            if($@) {
                $core->log(replace(gettext('Failed to load DBD::$DRIVER'), { DRIVER => $arg->{driver} }), CRITICAL)
            } else {
                $core->{_DBI}->{$arg->{driver}}->{DBD_LOADED} = 1;
            }
        } elsif($arg->{driver} eq 'mysql') {
            eval { require DBD::mysql; };
            if($@) {
                $core->log(replace(gettext('Failed to load DBD::$DRIVER'), { DRIVER => $arg->{driver} }), CRITICAL)
            } else {
                $core->{_DBI}->{$arg->{driver}}->{DBD_LOADED} = 1;
            }
        } else {
        #SQLite
        }
    }

    $arg->{name} ||= 'dbh';

    $arg->{options} ||= {
    #PrintError => 0, # warn() on errors
    #RaiseError => 0, # don't die on error
        PrintError => 1, # warn() on errors
            RaiseError => 1, # don't die on error
            AutoCommit => 1, # commit executes immediately
    };

    my $dbh;
    if($arg->{driver} eq 'Pg') {
        if($dbh = DBI->connect('DBI:Pg:dbname=' . $arg->{database}, $arg->{username}, $arg->{password}, $arg->{options})) {
            $core->log(replace(gettext('Connected to database "$DATABASE". ($NAME)'), { DATABASE => $arg->{database}, NAME => $arg->{name}}), INFO);
        } else {
            $core->log(replace(gettext('Couldn\'t connect to database "$DATABASE". ($NAME)'), { DATABASE => $arg->{database}, NAME => $arg->{name}}), INFO);
            return 0;
        }
    } elsif($arg->{driver} eq 'mysql') {
        if($dbh = DBI->connect('DBI:mysql:database=' . $arg->{database}, $arg->{username}, $arg->{password}, $arg->{options})) {
            $core->log(replace(gettext('Connected to database "$DATABASE". ($NAME)'), { DATABASE => $arg->{database}, NAME => $arg->{name}}), INFO);
        } else {
            $core->log(replace(gettext('Couldn\'t connect to database "$DATABASE". ($NAME)'), { DATABASE => $arg->{database}, NAME => $arg->{name}}), INFO);
            return 0;
        }
    } else {
    #SQLite
        my $filename = sprintf("%s/%s.db", $core->config->{DataDir}, $arg->{database});
        if($dbh = DBI->connect('DBI:SQLite:dbname=' . $filename, undef, undef, $arg->{options})) {
            $core->log(replace(gettext('Connected to database "$DATABASE". ($NAME)'), { DATABASE => $arg->{database}, NAME => $arg->{name}}), INFO);
        } else {
            $core->log(replace(gettext('Couldn\'t connect to database "$DATABASE". ($NAME)'), { DATABASE => $arg->{database}, NAME => $arg->{name}}), INFO);
            return 0;
        }
    }
    $core->{_DBI}->{_dbh}->{$arg->{name}} = $dbh;
    return 1;
}

=head2 isInfo

Override antiflood.

=cut

sub isInfo
{
    my $core = shift;
    $core->heap->{floodoverride} = 1;
}

=head2 dbh

$dbh = $core->dbh($name);

If no name is given, the default dbh is returned.

=cut

sub dbh
{
    my $core = shift;
    my $name = shift;
    my $dbh = shift;

    $name ||= 'dbh';

    if($dbh) {
        $core->{_DBI}->{_dbh}->{$name} = $dbh;
    }
    return $core->{_DBI}->{_dbh}->{$name};
}

=head2 load_config($core, $file, $config)

    Used by Core to reload configuration files in a safe manner.

=cut

sub load_config
{
    my ($core, $file, $config) = @_;

    return unless($file);
    delete $INC{$config}; #Reload from disk.
        eval {
            require $file;
        };

    if($@) {
        if($core) {
            $core->log(replace(gettext('Couldn\'t reload configuration from $FILE.'), { FILE => $file }), ERROR);
        } else {
            print replace(gettext('Couldn\'t reload configuration from $FILE.'), { FILE => $file }) . "\n$@\n";
        }

        return; #Don't reload.

    } else {
        if($core) {
            foreach my $key (keys %config) {
                next if($key =~ /^(Regexes|Modules)$/);
                $core->config->{$key} = $config{$key};
            }
            if(@{ $config{Regexes} }) {
                push(@{ $core->config->{Regexes} }, @{ $config{Regexes} });
            }
            if(%{ $config{Modules} }) {
                foreach my $key (keys %{ $config{Modules} }) {
                    $core->config->{Modules}->{$key} = $config{Modules}->{$key};
                }
            }
        } else {
            foreach my $key (keys %config) {
                next if($key =~ /^(Regexes|Modules)$/);
                $config->{$key} = $config{$key};
            }
            if(@{ $config{Regexes} }) {
                push(@{ $config->{Regexes} }, @{ $config{Regexes} });
            }
            if(%{ $config{Modules} }) {
                foreach my $key (keys %{ $config{Modules} }) {
                    $config->{Modules}->{$key} = $config{Modules}->{$key};
                }
            }
        }
    }
}

=head2 module_config($core, \%config, $where)

Load module configuration with respect to overrides.

Example from conf.pl:

Google => {
    '#channel' => {
        Results => 3,
    },
        GoogleKey => 'google_key',
        Results => 5,
}

Example in Google.pm:

my %config;
$core->module_config(\%config, $where[, $caller]); # $caller defaults to caller().

If $where is '#channel', $config{Results} will be 3. 
$config{GoogleKey} remains the same. For every other channels or in privmsg, $config{Results} will be 5.

=cut

sub module_config
{
    my $core = shift;
    my ($ref, $where, $caller) = @_;

    $caller ||= caller();

    my $config = $core->config->{Modules}->{$caller};
    return unless(ref($config) eq 'HASH');

    foreach my $key (keys %{$config}) {
        if($key =~ /^[#&]/ || $key eq 'privmsg') {
            next;
        } else {
            $ref->{$key} = $config->{$key};
        }
    }

    #Look for overrides.
    if($where && $config->{$where} && ref($config->{$where}) eq 'HASH') {
        foreach my $key (keys %{$config->{$where}}) {
            $ref->{$key} = $config->{$where}->{$key};
        }
    }

}

sub Reload
{
    my $core = shift;
    return if($core->testing);

    $core->log(replace(gettext('Rehash requested by "$WHO".'), { WHO => $core->heap->{who} }), NORMAL);
    $core->Rehash;
    return;
}


sub startConversation
{
    my $core = shift;
    my ($nick, $next, $question, $r) = @_;

    return -1 if($nick =~ /^[#&]/); #Can't conversate with channel.

        $r = {} unless(ref($r) eq 'HASH');

    $r->{started} = time(); $r->{module} = scalar(caller()); $r->{next} = $next;
    $core->var({ namespace => 'conversation', key => $core->heap->{nick}, value => 'fitte'});
    $core->{conversation}->{$nick} = $r;
    $core->msg($nick, $question, { flood => 1 });
}

sub conversate 
{
    my $core = shift;
    my $msg = shift;
    my $r = $core->{conversation}->{$core->heap->{nick}};

    my $next = $r->{next};
    if($next) {
        $core->isInfo;

        eval {
            $next = &{$r->{module} . "::$next"}($core, $r, $msg);
        };
        if($@) {
            $core->log("conversate: " . $@, ERROR);
            delete $core->conversation->{$core->heap->{nick}}; #Clean up.
                return 0;
        }
    }

    if(defined($next)) {
        $r->{next} = $next;
        return 1;
    } else {
        $r->{next} = undef;
        $core->{conversation}->{$core->heap->{nick}} = {};
        delete $core->{conversation}->{$core->heap->{nick}};
        return 0;
    }
}

sub cmsg { 
    my $core = shift;
    $core->msg($core->heap->{nick}, @_);
}

sub where { return shift->heap->{where}; }

=head1 SEE ALSO

Plugin.pm

=cut

=head1 AUTHOR

Alex Brasetvik < alex at brasetvik dot com >
Idar Lund < effbiai at users dot sf dot net >

=cut
1;
