#
# DBI extension layer
#
package My::DBI;
require Exporter;
use Carp;
use DBI;
use Exporter;
use IO::File;
use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
use warnings;

@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw
    (
     prepare_and_exec
     dbi_exec_for_column_list
     dbi_exec_for_hoh
     dbi_exec_for_hol
     dbi_exec_for_loh
     dbi_exec_for_lol
     dbi_exec_for_paired_hash
     dbi_exec_for_row_hash
     dbi_exec_for_row_list
     dbi_exec_for_scalar
     dbi_exists
     dbi_fetch_column_list
     dbi_fetch_hoh
     dbi_fetch_hol
     dbi_fetch_loh
     dbi_fetch_lol
     dbi_fetch_paired_hash
     dbi_fetch_row_hash
     dbi_fetch_row_list
     dbi_fetch_scalar
     dbi_insert_from_hash
     dbi_member_cond
     dbi_save_from_hash
     dbi_update_from_hash
     DBI_CACHE_STMT
     );
%EXPORT_TAGS =
    (
     'all'    => [@EXPORT_OK],
     'exec'   => [grep(/^dbi_exec/, @EXPORT_OK)],
     'fetch'  => [grep(/^dbi_fetch/, @EXPORT_OK)],
     'misc'   => [qw(dbi_exists dbi_insert_from_hash dbi_member_cond dbi_save_from_hash dbi_update_from_hash)]
     );

# Flag constants
sub DBI_CACHE_STMT { 0 }

sub prepare_and_exec {
    my $dbh = shift;
    my $flags = 0;
    $flags = shift if ($_[0] =~ /^\d+$/);
    my ($query, @binds) = @_;
    my $sth = (($flags & DBI_CACHE_STMT) ? $dbh->prepare_cached($query) : $dbh->prepare($query))
	or croak "error preparing query: ".$dbh->errstr." ($query)";
    $sth->execute(@binds) or croak "error executing query: ".$dbh->errstr." ($query)";
    return $sth;
}

sub dbi_exec_for_column_list { dbi_fetch_column_list (prepare_and_exec(@_)) }
sub dbi_exec_for_hoh { dbi_fetch_hoh (prepare_and_exec(@_)) }
sub dbi_exec_for_hol { dbi_fetch_hol (prepare_and_exec(@_)) }
sub dbi_exec_for_loh { dbi_fetch_loh (prepare_and_exec(@_)) }
sub dbi_exec_for_lol { dbi_fetch_lol (prepare_and_exec(@_)) }
sub dbi_exec_for_paired_hash { dbi_fetch_paired_hash (prepare_and_exec(@_)) }
sub dbi_exec_for_row_hash { dbi_fetch_row_hash (prepare_and_exec(@_)) }
sub dbi_exec_for_row_list { dbi_fetch_row_list (prepare_and_exec(@_))  }
sub dbi_exec_for_scalar { dbi_fetch_scalar (prepare_and_exec(@_)) }

sub dbi_exists {
    my $sth = prepare_and_exec(@_);
    return defined($sth->fetchrow_arrayref);
}

sub dbi_fetch_column_list {
    my ($sth) = shift;
    my $lref = $sth->fetchall_arrayref([0]);
    return map($_->[0], @$lref);
}

sub dbi_fetch_hoh {
    my ($sth) = shift;
    my $first_key = $sth->{NAME}->[0];
    my %hoh;
    while (my $href = $sth->fetchrow_hashref) {
	my $first_val = $href->{$first_key};
	$hoh{$first_val} = $href;
    }
    return %hoh;
}

sub dbi_fetch_hol {
    my ($sth) = shift;
    my $first_key = $sth->{NAME}->[0];
    my %hol;
    while (my $lref = $sth->fetchrow_arrayref) {
	my $first_val = shift(@$lref);
	$hol{$first_val} = $lref;
    }
    return %hol;
}

sub dbi_fetch_loh {
    my ($sth) = shift;
    my $lref = $sth->fetchall_arrayref({});
    return @$lref;
}

sub dbi_fetch_lol {
    my ($sth) = shift;
    my $lref = $sth->fetchall_arrayref([]);
    return @$lref;
}

sub dbi_fetch_paired_hash {
    my ($sth) = shift;
    die "query asks for ".$sth->{NUM_OF_FIELDS}." fields, not 2" if $sth->{NUM_OF_FIELDS} != 2;
    my $first_key = $sth->{NAME}->[0];
    my %h;
    while (my $lref = $sth->fetchrow_arrayref) {
	my $first_val = $lref->[0];
	$h{$first_val} = $lref->[1];
    }
    return %h;
}

sub dbi_fetch_row_hash {
    my ($sth) = shift;
    my $href = $sth->fetchrow_hashref;
    return ($href) ? %$href : ();
}

sub dbi_fetch_row_list {
    my ($sth) = shift;
    return $sth->fetchrow_array;
}

sub dbi_fetch_scalar {
    my ($sth) = shift;
    die "query asks for ".$sth->{NUM_OF_FIELDS}." fields, not 1" if $sth->{NUM_OF_FIELDS} != 1;
    my $lref = $sth->fetchrow_arrayref;
    return undef if !defined($lref);
    die "query returned more than one value" if (defined($sth->fetchrow_arrayref));
    return $lref->[0];
}

sub dbi_member_cond {
    my ($fieldName, $lref, $chunk) = @_;

    $chunk ||= 100;
    my @clauses;
    my @items = @$lref;
    return '(0=1)' if (!@items);
    return "$fieldName=$items[0]" if (@items==1);
    while (@items) {
	my @nextItems = splice(@items, 0, $chunk);
	push (@clauses, "$fieldName in (" . join(", ", @nextItems) . ")");
    }
    return "(".join("\nor\n", @clauses).")";
}

sub dbi_save_from_hash {
    my ($dbh, $table, $primaryKey, $rowRef) = @_;
    my %row = %$rowRef;
    if (exists($row{$primaryKey}) and dbi_exists($dbh, "select $primaryKey from $table where $primaryKey = ?", $row{$primaryKey})) {
	my $primaryValue = $row{$primaryKey};
	delete($row{$primaryKey});
	dbi_update_from_hash($dbh, $table, $rowRef, "$primaryKey = ?", $primaryValue);
    } else {
	dbi_insert_from_hash($dbh, $table, $rowRef);
    }
}

sub dbi_insert_from_hash {
    my ($dbh, $table, $rowRef) = @_;
    my %row = %$rowRef;
    my $fieldList = join(", ", keys(%row));
    my $bindList = join(", ", split(//, '?' x keys(%row)));
    my $query = "insert into $table ($fieldList) values ($bindList)";
    prepare_and_exec($dbh, $query, values(%row));
}

sub dbi_update_from_hash {
    my ($dbh, $table, $rowRef, $where, @params) = @_;
    my %row = %$rowRef;
    my $setList = join(", ", map("$_=?", keys(%row)));
    my $query = "update $table set $setList where $where";
    prepare_and_exec($dbh, $query, values(%row), @params);
}

__END__

=head1 NAME

MHQ::DBI - database utilities

=head1 SYNOPSIS

 use MHQ::DBI qw(:all);

 my $name = dbi_exec_for_scalar($dbh, "select name from employees where id = ?", $id);
 my %emp = dbi_exec_for_row_hash($dbh, DBI_CACHE_STMT, "select * from employees where id = ?", $id);
 my @emps = dbi_exec_for_loh($dbh, "select * from employees");
 ...

 dbi_save_from_hash($dbh, "employees", "id", %emp);

=head1 DESCRIPTION

This module provides convenience methods on top of DBI to execute
statements into a wide variety of data structures and perform other
common database tasks. These methods maintain database portability and
can be mixed freely with standard DBI methods.

=head1 EXECUTE METHODS

The following methods combine the preparation, execution, and fetching
of a statement in one convenient package. The methods are
distinguished by whether they select one or all rows, and what kind of
data structure they return.

When a method returns a row as a hash, the keys are the lowercase
field names.  This is in contrast to DBI's fetchrow_hashref(), which may
use either case depending on the underlying database.

Any DBI errors that occur during preparation or execution
automatically result in a die(). You can wrap these methods in eval()
if you want to detect and handle errors.

Each method takes the same parameters: 

    ($dbh, [$flags], $select, [@bind_vars])

  $dbh: a valid database handle
  $flags: optionally one or more numeric flags OR'd together (see below)
  $select: a select statement
  @binds: optionally one or more bind variables corresponding to placeholders in the select

The possible flags are:

C<DB_CACHE_STMT>: Indicates that prepare_cached should be used instead of prepare,
which means that the statement handler will be cached and reused every time the
identical SQL statement occurs within this database handle. If you use this, be
sure the statement remains exactly the same (i.e. use bind variables)!

Here are the methods:

=over

=item dbi_exec_for_column_list

Selects all rows; returns the first column's contents from each row as
a list.

=item dbi_exec_for_hoh

Selects all rows; returns the data as a hash of hashes, where the key
is the first column's contents and the value is the entire row
(including the first column) as another hash.

=item dbi_exec_for_hol

Selects all rows; returns the data as a hash of lists, where the key
is the first column's contents and the value is the rest of the row
(not including the first column) as a list.

=item dbi_exec_for_loh

Selects all rows; returns the data as a list of hashes, where each
element represents one row as a hash.

=item dbi_exec_for_lol

Selects all rows; returns the data as a list of lists, where each
element represents one row as a list.

=item dbi_exec_for_paired_hash

Selects all rows; returns the data as a hash, where the key is the
first column's contents and the value is the second column's contents.

=item dbi_exec_for_row_hash

Selects the first row; returns the row as a hash.

=item dbi_exec_for_row_list

Selects the first row; returns the row as a list.

=item dbi_exec_for_scalar

Selects the first row; returns the first column's contents as a
scalar.

=item dbi_exists

Attempts to select the first row; returns true if a row was found, false otherwise.

=back

=head1 FETCH METHODS

The following methods are exactly analogous to the execute methods
above, except that they take a statement handle that has already been
prepared and executed.

      dbi_fetch_column_list
      dbi_fetch_hoh
      dbi_fetch_hol
      dbi_fetch_loh
      dbi_fetch_lol
      dbi_fetch_paired_hash
      dbi_fetch_row_hash
      dbi_fetch_row_list
      dbi_fetch_scalar

=head1 MISCELLANEOUS METHODS

=over

=item dbi_member_cond ($field, $values, [$chunk_size])

Given a field name ($field) and a reference to a list of values
($values), this method constructs and returns a membership condition
of the form

    (field in (value1, value2, ...))

Every database system has some limit on the number of values in a set.
Thus, dbi_member_cond will break up the conditions into multiple or
clauses if the number of values is greater than 100. e.g. given 150
$values,

    (field in (value1, value2, ..., value100) or field in (value101, value102, ...value150))
 
The chunk size can be overridden with the optional $chunk_size
parameter.

=item dbi_save_from_hash ($dbh, $table, $primary_key, $row)

Given a valid database handle ($dbh), a table name ($table), a primary
key name ($primary_key), and a reference to a hash representing a row
($row), this method saves the row in the table.  If a row with the
same primary key value already exists, dbi_save_from_hash performs an
update; otherwise it performs an insert.

=item dbi_insert_from_hash ($dbh, $table, $row)

Given a valid database handle ($dbh), a table name ($table), and a
reference to a hash representing a row ($row), this method inserts the
row in the table.

=item dbi_update_from_hash ($dbh, $table, $row, $where, @params)

Given a valid database handle ($dbh), a table name ($table), a
reference to a hash representing a full or partial row ($row), a where
clause ($where), and a set of optional bind params for the where
clause (@params), this method updates the matching row(s) in the
table.

=back

=head1 AUTHOR

Jonathan Swartz, swartz@transbay.net

October 5, 1999

=cut


